/*
 *  Copyright (c) 2003-2006, Shoichi Hasegawa and Springhead development team 
 *  All rights reserved.
 *  This software is free software. You can freely use, distribute and modify this 
 *  software. Please deal with this software under one of the following licenses: 
 *  This license itself, Boost Software License, The MIT License, The BSD License.   
 */
#include "Physics.h"
#ifdef USE_HDRSTOP
#pragma hdrstop
#endif
#include <float.h>

using namespace PTM;
namespace Spr{

///////////////////////////////////////////////////////////////////
//	PHSolid
IF_OBJECT_IMP(PHSolid, SceneObject);

PHSolid::PHSolid(const PHSolidDesc& desc, SceneIf* s):PHSolidDesc(desc){
	bDynamical = true;
	integrationMode = PHINT_SIMPLETIC;
	inertia_inv = inertia.inv();
	if (s){ SetScene(s); }
}
void PHSolid::SetScene(SceneIf* s){
	SceneObject::SetScene(s);
	SetGravity(gravity);
}
CDShapeIf* PHSolid::CreateShape(const CDShapeDesc& desc){
	CDShapeIf* rv = DCAST(PHSceneIf, GetScene())->CreateShape(desc);
	if (rv){
		AddShape(rv);
	}
	return rv;
}
ObjectIf* PHSolid::CreateObject(const IfInfo* info, const void* desc){
	ObjectIf* rv = SceneObject::CreateObject(info, desc);
	if (!rv){
		if (info->Inherit(CDShapeIf::GetIfInfoStatic())){
			rv = CreateShape(*(CDShapeDesc*)desc);
		}
	}
	return rv;
}
bool PHSolid::AddChildObject(ObjectIf* obj){
	if (DCAST(CDShapeIf, obj)){
		AddShape(DCAST(CDShapeIf, obj));
		return true;
	}
	return false;
}

void PHSolid::CalcBBox(){
	Vec3f bboxMin = Vec3f(FLT_MAX, FLT_MAX, FLT_MAX);
	Vec3f bboxMax = Vec3f(-FLT_MAX,-FLT_MAX,-FLT_MAX);
	for(int i = 0; i < (int)shapes.size(); i++){
		shapes[i].shape->CalcBBox(bboxMin, bboxMax, shapes[i].pose);
	}
	if (bboxMin.X() == FLT_MAX){
		bbox.SetBBoxMinMax(Vec3f(0,0,0), Vec3f(-1,-1,-1));
	}else{
		bbox.SetBBoxMinMax(bboxMin, bboxMax);
	}
//	DSTR << GetName() << bbox.GetBBoxMin() << bbox.GetBBoxMax() << std::endl;
}
void PHSolid::GetBBoxSupport(const Vec3f& dir, float& minS, float& maxS){
		bbox.GetSupport(GetOrientation().Inv()*dir, minS, maxS);
		float c = Vec3f(GetFramePosition()) * dir;
		minS += c;
		maxS += c;
}

void PHSolid::Step(){
	force = nextForce;
	torque = nextTorque;
	nextForce.clear();
	nextTorque.clear();

	//ɑ̃GWɂčXVꂽꍇ͐ϕsȂ
	if(IsUpdated())
		return;

	PHScene* s = DCAST(PHScene, GetScene());
	double dt = s->GetTimeStep();
	assert(GetIntegrationMode() != PHINT_NONE);
#ifdef _DEBUG
	if (!finite(velocity.norm()) || velocity.norm() > 100 || angVelocity.norm() > 100){	
		DSTR << "Warning: solid '" << GetName() << "' has a very fast velocity. v:" << velocity << "w:" << angVelocity << std::endl;
	}
#endif
	//k1 = f(y);
	//k2 = f(y + k1 * h / 2);
	//k3 = f(y + k2 * h / 2);
	//k4 = f(y + k3);
	//y += (k1 + 2 * k2 + 2 * k3 + k4) * h / 6;
	//	ϕvZ
	Vec3d dv, dw;				//<	xEpx̕ω
	Vec3d	_angvel[4];			//<	lϕW
	Vec3d	_angacc[4];
	switch(GetIntegrationMode()){
	case PHINT_EULER:
		//sړʂ̐ϕ
		SetCenterPosition(GetCenterPosition() + velocity * dt);
		velocity += force * (dt / mass);
		//pxNEH[^jI̎Ԕ߁AϕAK
		pose.Ori() += pose.Ori().Derivative(angVelocity) * dt;
		pose.Ori().unitize();
		torque		= pose.Ori().Conjugated() * torque;			//gNƊpx[JW
		angVelocity = pose.Ori().Conjugated() * angVelocity;
		angVelocity += Euler(inertia, torque, angVelocity) * dt;	//IC[̉^
		torque = pose.Ori() * torque;						//gNƊpx[h
		angVelocity = pose.Ori() * angVelocity;
		break;
	case PHINT_ARISTOTELIAN:{
		SetCenterPosition(GetCenterPosition() + velocity * dt);
		velocity = force / mass;		//x͗͂ɔႷ
		Vec3d tq = pose.Ori().Conjugated() * torque;	//gN[J
		angVelocity = pose.Ori() * (inertia_inv * tq);	//px̓gNɔႷ
		//NEH[^jIϕAK
		pose.Ori() += pose.Ori().Derivative(angVelocity) * dt;
		pose.Ori().unitize();
		}break;
	case PHINT_SIMPLETIC:{
		//	x(dt) = x(0) + dt*v(0)/m
		//	v(dt) = v(0) + dt*f(dt)
		//]ʂ̐ϕ
		torque		= pose.Ori().Conjugated() * torque;				//	gNƊpx[JW
		angVelocity = pose.Ori().Conjugated() * angVelocity;

		dw = Euler(inertia, torque, angVelocity) * dt;			//pxω
		angVelocity += dw;										//px̐ϕ
		Quaterniond dq = Quaterniond::Rot(angVelocity * dt);
		Vec3d dp = pose.Ori() * (dq*(-center) - (-center));
		pose.Ori() = pose.Ori() * dq;
		pose.Ori().unitize();
		torque = pose.Ori() * torque;									//gNƊpx[h
		angVelocity = pose.Ori() * angVelocity;
		//sړʂ̐ϕ
		velocity += force * (dt / mass);								//	x̐ϕ
		SetCenterPosition(GetCenterPosition() + velocity * dt + dp);	//	ʒu̐ϕ
		}break;
	case PHINT_ANALYTIC:{
		//]ʂ̐ϕ
		//]͉͓IɐϕłȂ̂ŁA`IɁ̌]̏ꍇɓĂ͂߂
		torque		= pose.Ori().Conjugated() * torque;					//gNƊpx[JW
		angVelocity = pose.Ori().Conjugated() * angVelocity;
		dw = Euler(inertia, torque, angVelocity) * dt;			//pxω

		Quaterniond dq = Quaterniond::Rot((angVelocity+0.5*dw) * dt);
		Vec3d dp = pose.Ori() * (dq*(-center) - (-center));
		pose.Ori() = pose.Ori() * dq;
		pose.Ori().unitize();

		angVelocity += dw;										//px̐ϕ
		torque = pose.Ori() * torque;									//gNƊpx[h
		angVelocity = pose.Ori() * angVelocity;
		//sړʂ̐ϕi͉Ɉvj
		dv = force * (dt / mass);									//xω
		SetCenterPosition(GetCenterPosition() + (velocity+0.5*dv) * dt + dp);	//	ʒu̐ϕ
		velocity += dv;												//x̐ϕ
		}break;
	case PHINT_RUNGEKUTTA2:
		//sړʂ̐ϕ(܂߂ɃQNb^ɏ]Ăʂ͓͉)
		dv = force * (dt / mass);
		SetCenterPosition(GetCenterPosition() + (velocity + 0.5 * dv) * dt);
		velocity += dv;
		//]ʂ̌vZ
		//]͉͓IɐϕłȂ̂ŁAQNb^g
		torque		= pose.Ori().Conjugated() * torque;					//gNƊpx[JW
		angVelocity = pose.Ori().Conjugated() * angVelocity;
		_angvel[0]	= angVelocity;
		_angacc[0]	= Euler(inertia, torque, _angvel[0]);
		_angvel[1]	= _angvel[0] + _angacc[0] * dt;
		_angacc[1]	= Euler(inertia, torque, _angvel[1]);
		pose.Ori() += pose.Ori().Derivative(pose.Ori() * (_angvel[0] + _angvel[1]) / 2.0) * dt;
		pose.Ori().unitize();
		angVelocity = pose.Ori() * (angVelocity + ((_angacc[0] + _angacc[1]) * (dt / 2.0)));
		torque = pose.Ori() * torque;
		break;
	case PHINT_RUNGEKUTTA4:
		//sړʂ̐ϕ(܂߂ɃQNb^ɏ]Ăʂ͓͉)
		dv = force * (dt / mass);
		SetCenterPosition(GetCenterPosition() + (velocity + 0.5 * dv) * dt);
		velocity += dv;
		//]ʂ̌vZ
		_angvel[0]	= angVelocity;
		_angacc[0]	= Euler(inertia, torque, _angvel[0]);
		_angvel[1]	= _angvel[0] + _angacc[0] * (dt / 2.0);
		_angacc[1]	= Euler(inertia, torque, _angvel[1]);
		_angvel[2]	= _angvel[0] + _angacc[1] * (dt / 2.0); 
		_angacc[2]	= Euler(inertia, torque, _angvel[2]);
		_angvel[3]	= _angvel[0] + _angacc[2] * dt;
		_angacc[3]	= Euler(inertia, torque, _angvel[3]);
		pose.Ori() += pose.Ori().Derivative(pose.Ori() * (_angvel[0] + 2.0 * (_angvel[1] + _angvel[2]) + _angvel[3]) / 6.0) * dt;
		pose.Ori().unitize();
		angVelocity = pose.Ori() * (angVelocity + (_angacc[0] + 2.0 * (_angacc[1] + _angacc[2]) + _angacc[3]) * (dt / 6.0));
		torque = pose.Ori() * torque;
		break;
	default:			
		/* DO NOTHING */		
		break;
	}
}
void PHSolid::AddTorque(Vec3d t){
	nextTorque += t; 
}
void PHSolid::AddForce(Vec3d f){
	nextForce += f;
}

void PHSolid::AddForce(Vec3d f, Vec3d r){
	nextTorque += (r - pose*center) ^ f;
	nextForce += f;
}

/*void PHSolid::AddForceLocal(Vec3d f, Vec3d r){
	torque += cross(pose.Ori() * (r - center), f);
	force += f;
}*/


void PHSolid::AddShape(CDShapeIf* shape){
	CDShape* sh = DCAST(CDShape, shape);
	shapes.push_back(CDShapeRefWithPose());
	shapes.back().shape = sh;
	CalcBBox();
	//ڐGGWshapePairsXV
	PHScene* scene = DCAST(PHScene,GetScene());
	scene->penaltyEngine->UpdateShapePairs(this);
	scene->constraintEngine->UpdateShapePairs(this);
}

Posed	PHSolid::GetShapePose(int i){
	if(0 <= i && i < (int)shapes.size())
		return shapes[i].pose;
	return Posed();
}

void	PHSolid::SetShapePose(int i, const Posed& pose){
	if(0 <= i && i < (int)shapes.size()){
		shapes[i].pose = pose;
		CalcBBox();
	}
}

void PHSolid::SetGravity(bool bOn){
	PHScene* ps = DCAST(PHScene,GetScene());
	PHGravityEngine* ge;
	ps->engines.Find(ge);
	if (bOn == true){
		if (!ge->solids.Find(this)){
			ge->solids.push_back(this);
		}
	}else{
		ge->solids.Erase(this);
	}
	gravity = bOn;
}
int PHSolid::NShape(){
	return shapes.size();
}
CDShapeIf* PHSolid::GetShape(int i){
	return shapes[i].shape;
}

//----------------------------------------------------------------------------
//	PHSolidContainer
//
OBJECT_IMP(PHSolidContainer, PHEngine);

PHSolidContainer::PHSolidContainer(){
}

bool PHSolidContainer::AddChildObject(ObjectIf* o){
	PHSolid* s = DCAST(PHSolid, o);
	if (s && std::find(solids.begin(), solids.end(), s) == solids.end()){
		solids.push_back(s);
		return true;
	}
	return false;
}
bool PHSolidContainer::DelChildObject(ObjectIf* o){
	PHSolid* so = DCAST(PHSolid, o);
	if (so){
		solids.Erase(so);
		return true;
	}
	return false;
}

void PHSolidContainer::Reset(){
	for(PHSolids::iterator it = solids.begin(); it != solids.end(); ++it)
		(*it)->SetUpdated(false);
}

void PHSolidContainer::Step(){
	for(PHSolids::iterator it = solids.begin(); it != solids.end(); ++it)
		(*it)->Step();
}

//----------------------------------------------------------------------------
//	PHSolidInitializer
//
OBJECT_IMP(PHSolidInitializer, PHEngine);
void PHSolidInitializer::Step(){
		container->Reset();	
}

}
